% Testing problem: Image deblurring (Barbara deblurring).
%
% Matlab software for the paper "The regularizing effect of the Golub-Kahan 
% iterative bidiagonalization and revealing the noise level in the data", 
% by I. Hnetynkova, M. Plesinger, and Z. Strakos.
%
% Amult.m ......... Matrix-vector product y = A * x via 2D convolution.
% ATmult.m ........ Matrix-vector product y = A^T * x via 2D convolution.
% Barbara_256.m ... Testing image Barbara, 256 x 256 pixels, gray-scale.
% Barbara_512.m ... Testing image Barbara, 512 x 512 pixels, gray-scale 
%                   (further standard testing images are available in
%                   the folder 'Images').
% Content.m ....... Table of content.
% csvd.m .......... Compact singular value decomposition, part of [1].
% goka_b_PSF.m .... Golub-Kahan bidiagonalization using matrix-vector products
%                   computed via 2D convolution; it is a modified routine 
%                   lanc_b.m from [1]. (Note that, if the bidiagonaliztion is
%                   computed with Householder reorthogonalization, the routine
%                   app_hh.m from [1] is needed, it is not a part of this pack.)
% psfGauss.m ...... Contstructs a Gaussian blur point-spread-function, 
%                   part of [2].
% testDeblur_256.m  Launches a smaller testing problem!
% testDeblur_512.m  Launches a larger testing problem!
%
% [1] P. C. Hansen -- Regularization Tools, Version 4.0 for MATLAB 7.3.
% [2] P. C. Hansen, J. Nagy, D. O'Leary -- Deblurring Images: Matrices, 
%     Spectra, and Filtering, SIAM, 2006 (additional software tools).

% Author of this code: P. C. Hansen; modified by M. Plesinger, April, 2009.

clear;

% Number of iterations of the Golub-Kahan bidiagonalization ...
kmax = 60;  

% ... without reorthogonalization.
reorth = 0;

% Load the testing image.
fprintf('\nLoading Barbara_512 image ........................');
X_exact = imread('Barbara_512.bmp');
X_exact = double(X_exact);
fprintf(' OK\n');

% The image dimensions.
[M,N] = size(X_exact);

% The problem dimensions.
m = M*N; n = m;

% The exact solution.
x_exact = X_exact(:);

% Constructs the Gaussian blur point-spread-function (PSF).
% Works only when the size "dim" is odd!!!
fprintf('Constructing the point-spread-function ...........');
dim = 11;  
sigma = 3;
PSF = psfGauss(dim,sigma);
fprintf(' OK\n');

% The exact right-hand side computed via 2D convolution.
fprintf('Constructing the noisy right-hand side ...........');
B_exact = conv2(X_exact,PSF,'same');
b_exact = B_exact(:);

% The additional white-noise and the noisy right-hand side.
delta_noise = 3e-3;
b_noise = randn(m,1); 
b_noise = b_noise/norm(b_noise);
b_noise = norm(b_exact) * b_noise * delta_noise;
b = b_exact + b_noise;
beta = norm(b);
fprintf(' OK\n');

% Run the Golub-Kahan bidiagonalization.
fprintf('Computing the Golub-Kahan bidiagonalization ......');
[S,L,W] = goka_b_PSF(PSF,M,N,b,kmax,reorth);
fprintf(' OK\n');

% Initialization: noise level estimator, approx. soltuions, and error history. 
fprintf('Estimating the noise level, and computing\n');
fprintf('the LSQR approximations of the exact solution ....');
nl_estimate = zeros(kmax,1);
X = zeros(n,kmax);
err = zeros(kmax,1);

% Computes the SVD of L_k and approximates the solution, for k = 1,...,kmax.
for k = 1:kmax,

    % The SVD components has standard ordeding, here. Please note that in the
    % paper is used inverse ordering. Thus, we look for the first component
    % of the left singular vector corresponding to Theta(k,k) here.
    [P,Theta,Q] = csvd(L(1:k,1:k));
    nl_estimate(k) = P(1,k);

    % LSQR approximation of the solution, obtained using backslash operator!!!
    e1 = zeros(k+1,1); 
    e1(1) = 1;
    X(:,k) = beta*W(:,1:k)*(L(1:k+1,1:k)\e1);
    err(k,1) = norm(X(:,k)-x_exact);

end
[minerr,Idx] = min(err);
fprintf(' OK\n');

% Plotting graphs ...
fprintf('Plotting graphs ..................................');
figure(1); clf;

% ... of the estimator, ....
subplot(2,1,1);
h = semilogy(1:kmax,abs(nl_estimate),'k-o',[1,kmax],delta_noise*[1,1],'r-.');
set(h(1),'LineWidth',2);
title(['| ( p_1^{(k)} , e_1 ) |,   || b^{noise} ||_2 / || b^{exact} ||_2 = ', num2str(norm(b_noise)/norm(b_exact))]);
legend('| ( p_1^{(k)} , e_1 ) |','\delta_{noise}');

% ... and the error history.
subplot(2,1,2);
h = semilogy(1:kmax,err,'k-',Idx,err(Idx),'ro');
set(h(2),'LineWidth',2);
set(h(2),'MarkerSize',12);
axis([0,kmax,err(Idx)*0.9,max(err)/0.9]);
title('error history');
legend('|| x_k^{LSQR} - x^{exact} ||',sprintf('the minimal error, k = %d',Idx));

% Image reconstruction ...
figure(2); clf; 
colormap gray;

% ... exact solution, ...
subplot(2,2,1);
imagesc(X_exact); axis image off;
title('exact image');

% ... noisy right-hand side, ...
subplot(2,2,2);
imagesc(reshape(b,M,N)); axis image off;
title('noisy and blurred');

% ... the best LSQR reconstruciton, ...
subplot(2,2,3);
imagesc(reshape(X(:,Idx),M,N)); axis image off;
title('LSQR reconstruction with minimal error');

% ... and the point-spread-function.
subplot(2,2,4);
imagesc(PSF); axis image off;
title('Gaussian point-spread-function');
fprintf(' OK\n\n');

% Componentwise error ...
figure(3); clf; 
colormap hot;
subplot(1,1,1);
imagesc(abs(X_exact-reshape(X(:,Idx),M,N))); colorbar; axis image off;
title(sprintf('componentwise error | x_{%d}^{LSQR} - x^{exact} |',Idx));
